import os
from pathlib import Path
from typing import Literal

import pooch
from brainglobe_utils.general.config import get_config_obj

from cellfinder import DEFAULT_CELLFINDER_DIRECTORY
from cellfinder.core.tools.source_files import (
    default_configuration_path,
    user_specific_configuration_path,
)

DEFAULT_DOWNLOAD_DIRECTORY = DEFAULT_CELLFINDER_DIRECTORY / "models"


MODEL_URL = "https://gin.g-node.org/cellfinder/models/raw/master"

model_filenames = {
    "resnet50_tv": "resnet50_tv.h5",
    "resnet50_all": "resnet50_weights.h5",
}

model_hashes = {
    "resnet50_tv": "63d36af456640590ba6c896dc519f9f29861015084f4c40777a54c18c1fc4edd",  # noqa: E501
    "resnet50_all": None,
}


model_type = Literal["resnet50_tv", "resnet50_all"]


def download_models(
    model_name: model_type, download_path: os.PathLike
) -> Path:
    """
    For a given model name and download path, download the model file
    and return the path to the downloaded file.

    Parameters
    ----------
    model_name : model_type
        The name of the model to be downloaded.
    download_path : os.PathLike
        The path where the model file will be downloaded.

    Returns
    -------
    Path
        The path to the downloaded model file.

    """

    download_path = Path(download_path)
    filename = model_filenames[model_name]
    model_path = pooch.retrieve(
        url=f"{MODEL_URL}/{filename}",
        known_hash=model_hashes[model_name],
        path=download_path,
        fname=filename,
        progressbar=True,
    )

    return Path(model_path)


def amend_user_configuration(new_model_path=None) -> None:
    """
    Amends the user configuration to contain the configuration
    in new_model_path, if specified.

    Parameters
    ----------
    new_model_path : Path, optional
        The path to the new model configuration.
    """
    print("(Over-)writing custom user configuration")

    original_config = default_configuration_path()
    new_config = user_specific_configuration_path()
    if new_model_path is not None:
        write_model_to_config(new_model_path, original_config, new_config)


def write_model_to_config(new_model_path, orig_config, custom_config):
    """
    Update the model path in the custom configuration file, by
    reading the lines in the original configuration file, replacing
    the line starting with "model_path =" and writing these
    lines to the custom file.

    Parameters
    ----------
    new_model_path : str
        The new path to the model.
    orig_config : str
        The path to the original configuration file.
    custom_config : str
        The path to the custom configuration file to be created.

    Returns
    -------
    None

    """
    config_obj = get_config_obj(orig_config)
    model_conf = config_obj["model"]
    orig_path = model_conf["model_path"]

    with open(orig_config, "r") as in_conf:
        data = in_conf.readlines()
    for i, line in enumerate(data):
        data[i] = line.replace(
            f"model_path = '{orig_path}", f"model_path = '{new_model_path}"
        )

    custom_config_path = Path(custom_config)
    custom_config_path.parent.mkdir(parents=True, exist_ok=True)
    with open(custom_config, "w") as out_conf:
        out_conf.writelines(data)



from pathlib import Path

from cellfinder import DEFAULT_CELLFINDER_DIRECTORY


def default_configuration_path():
    """
    Returns the default configuration path for cellfinder.

    Returns:
        Path: The default configuration path.
    """
    return Path(__file__).parent.parent / "config" / "cellfinder.conf"


def user_specific_configuration_path():
    """
    Returns the path to the user-specific configuration file for cellfinder.

    This function returns the path to the user-specific configuration file
    for cellfinder. The user-specific configuration file is located in the
    user's home directory under the ".brainglobe/cellfinder" folder
    and is named "cellfinder.conf.custom".

    Returns:
        Path: The path to the custom configuration file.

    """
    return DEFAULT_CELLFINDER_DIRECTORY / "cellfinder.conf.custom"



from functools import partial
from math import ceil
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple

import napari
import napari.layers
from brainglobe_utils.cells.cells import Cell
from magicgui import magicgui
from magicgui.widgets import FunctionGui, ProgressBar
from napari.utils.notifications import show_info
from qtpy.QtWidgets import QScrollArea

from cellfinder.core.classify.cube_generator import get_cube_depth_min_max
from cellfinder.napari.utils import (
    add_classified_layers,
    add_single_layer,
    cellfinder_header,
    html_label_widget,
    napari_array_to_cells,
)

from .detect_containers import (
    ClassificationInputs,
    DataInputs,
    DetectionInputs,
    MiscInputs,
)
from .thread_worker import Worker

NETWORK_VOXEL_SIZES = [5, 1, 1]
CUBE_WIDTH = 50
CUBE_HEIGHT = 20
CUBE_DEPTH = 20

# If using ROI, how many extra planes to analyse
MIN_PLANES_ANALYSE = 0


def get_heavy_widgets(
    options: Dict[str, Any]
) -> Tuple[Callable, Callable, Callable]:
    # signal and other input are separated out from the main magicgui
    # parameter selections and are inserted as widget children in their own
    # sub-containers of the root. Because if these image parameters are
    # included in the root widget, every time *any* parameter updates, the gui
    # freezes for a bit likely because magicgui is processing something for
    # all the parameters when any parameter changes. And this processing takes
    # particularly long for image parameters. Placing them as sub-containers
    # alleviates this
    @magicgui(
        call_button=False,
        persist=False,
        scrollable=False,
        labels=False,
        auto_call=True,
    )
    def signal_image_opt(
        viewer: napari.Viewer,
        signal_image: napari.layers.Image,
    ):
        """
        magicgui widget for setting the signal_image parameter.

        Parameters
        ----------
        signal_image : napari.layers.Image
             Image layer containing the labelled cells
        """
        options["signal_image"] = signal_image
        options["viewer"] = viewer

    @magicgui(
        call_button=False,
        persist=False,
        scrollable=False,
        labels=False,
        auto_call=True,
    )
    def background_image_opt(
        background_image: napari.layers.Image,
    ):
        """
        magicgui widget for setting the background image parameter.

        Parameters
        ----------
        background_image : napari.layers.Image
             Image layer without labelled cells
        """
        options["background_image"] = background_image

    @magicgui(
        call_button=False,
        persist=False,
        scrollable=False,
        labels=False,
        auto_call=True,
    )
    def cell_layer_opt(
        cell_layer: napari.layers.Points,
    ):
        """
        magicgui widget for setting the cell layer input when detection is
        skipped.

        Parameters
        ----------
        cell_layer : napari.layers.Points
            If detection is skipped, select the cell layer containing the
            detected cells to use for classification
        """
        options["cell_layer"] = cell_layer

    return signal_image_opt, background_image_opt, cell_layer_opt


def add_heavy_widgets(
    root: FunctionGui,
    widgets: Tuple[FunctionGui, ...],
    new_names: Tuple[str, ...],
    insertions: Tuple[str, ...],
) -> None:
    for widget, new_name, insertion in zip(widgets, new_names, insertions):
        # make it look as if it's directly in the root container
        widget.margins = 0, 0, 0, 0
        # the parameters of these widgets are updated using `auto_call` only.
        # If False, magicgui passes these as args to root() when the root's
        # function runs. But that doesn't list them as args of its function
        widget.gui_only = True
        root.insert(root.index(insertion) + 1, widget)
        getattr(root, widget.name).label = new_name


def restore_options_defaults(widget: FunctionGui) -> None:
    """
    Restore default widget values.
    """
    defaults = {
        **DataInputs.defaults(),
        **DetectionInputs.defaults(),
        **ClassificationInputs.defaults(),
        **MiscInputs.defaults(),
    }
    for name, value in defaults.items():
        if value is not None:  # ignore fields with no default
            getattr(widget, name).value = value


def get_results_callback(
    skip_classification: bool, viewer: napari.Viewer
) -> Callable:
    """
    Returns the callback that is connected to output of the pipeline.
    It returns the detected points that we have to visualize.
    """
    if skip_classification:
        # after detection w/o classification, everything is unknown
        def done_func(points):
            add_single_layer(
                points,
                viewer=viewer,
                name="Cell candidates",
                cell_type=Cell.UNKNOWN,
            )

    else:
        # after classification we have either cell or unknown
        def done_func(points):
            add_classified_layers(
                points,
                viewer=viewer,
                unknown_name="Rejected",
                cell_name="Detected",
            )

    return done_func


def find_local_planes(
    viewer: napari.Viewer,
    voxel_size_z: float,
    signal_image: napari.layers.Image,
) -> Tuple[int, int]:
    """
    When detecting only locally, it returns the start and end planes to use.
    """
    current_plane = viewer.dims.current_step[0]

    # so a reasonable number of cells in the plane are detected
    planes_needed = MIN_PLANES_ANALYSE + int(
        ceil((CUBE_DEPTH * NETWORK_VOXEL_SIZES[0]) / voxel_size_z)
    )

    start_plane, end_plane = get_cube_depth_min_max(
        current_plane, planes_needed
    )
    start_plane = max(0, start_plane)
    end_plane = min(len(signal_image.data), end_plane)

    return start_plane, end_plane


def reraise(e: Exception) -> None:
    """Re-raises the exception."""
    raise Exception from e


def detect_widget() -> FunctionGui:
    """
    Create a detection plugin GUI.
    """
    progress_bar = ProgressBar()

    # options that is filled in from the gui
    options = {
        "signal_image": None,
        "background_image": None,
        "viewer": None,
        "cell_layer": None,
    }

    signal_image_opt, background_image_opt, cell_layer_opt = get_heavy_widgets(
        options
    )

    @magicgui(
        detection_label=html_label_widget("Cell detection", tag="h3"),
        **DataInputs.widget_representation(),
        **DetectionInputs.widget_representation(),
        **ClassificationInputs.widget_representation(),
        **MiscInputs.widget_representation(),
        call_button=True,
        persist=True,
        reset_button=dict(widget_type="PushButton", text="Reset defaults"),
        scrollable=True,
    )
    def widget(
        detection_label,
        data_options,
        voxel_size_z: float,
        voxel_size_y: float,
        voxel_size_x: float,
        detection_options,
        skip_detection: bool,
        soma_diameter: float,
        ball_xy_size: float,
        ball_z_size: float,
        ball_overlap_fraction: float,
        log_sigma_size: float,
        n_sds_above_mean_thresh: int,
        soma_spread_factor: float,
        max_cluster_size: int,
        classification_options,
        skip_classification: bool,
        trained_model: Optional[Path],
        use_pre_trained_weights: bool,
        misc_options,
        start_plane: int,
        end_plane: int,
        n_free_cpus: int,
        analyse_local: bool,
        debug: bool,
        reset_button,
    ) -> None:
        """
        Run detection and classification.

        Parameters
        ----------
        voxel_size_z : float
            Size of your voxels in the axial dimension
        voxel_size_y : float
            Size of your voxels in the y direction (top to bottom)
        voxel_size_x : float
            Size of your voxels in the x direction (left to right)
        skip_detection : bool
            If selected, the detection step is skipped and instead we get the
            detected cells from the cell layer below (from a previous
            detection run or import)
        soma_diameter : float
            The expected in-plane soma diameter (microns)
        ball_xy_size : float
            Elliptical morphological in-plane filter size (microns)
        ball_z_size : float
            Elliptical morphological axial filter size (microns)
        ball_overlap_fraction : float
            Fraction of the morphological filter needed to be filled
            to retain a voxel
        log_sigma_size : float
            Laplacian of Gaussian filter width (as a fraction of soma diameter)
        n_sds_above_mean_thresh : int
            Cell intensity threshold (as a multiple of noise above the mean)
        soma_spread_factor : float
            Cell spread factor (for splitting up cell clusters)
        max_cluster_size : int
            Largest putative cell cluster (in cubic um) where splitting
            should be attempted
        use_pre_trained_weights : bool
            Select to use pre-trained model weights
        skip_classification : bool
            If selected, the classification step is skipped and all cells from
            the detection stage are added
        trained_model : Optional[Path]
            Trained model file path (home directory (default) -> pretrained
            weights)
        start_plane : int
            First plane to process (to process a subset of the data)
        end_plane : int
            Last plane to process (to process a subset of the data)
        n_free_cpus : int
            How many CPU cores to leave free
        analyse_local : bool
            Only analyse planes around the current position
        debug : bool
            Increase logging
        reset_button :
            Reset parameters to default
        """
        # we must manually call so that the parameters of these functions are
        # initialized and updated. Because, if the images are open in napari
        # before we open cellfinder, then these functions may never be called,
        # even though the image filenames are shown properly in the parameters
        # in the gui. Likely auto_call doesn't make magicgui call the functions
        # in this circumstance, only if the parameters are updated once
        # cellfinder plugin is fully open and initialized
        signal_image_opt()
        background_image_opt()
        cell_layer_opt()

        signal_image = options["signal_image"]

        if signal_image is None or options["background_image"] is None:
            show_info("Both signal and background images must be specified.")
            return

        detected_cells = []
        if skip_detection:
            if options["cell_layer"] is None:
                show_info(
                    "Skip detection selected, but no existing cell layer "
                    "is selected."
                )
                return

            # set cells as unknown so that classification will process them
            detected_cells = napari_array_to_cells(
                options["cell_layer"], Cell.UNKNOWN
            )

        data_inputs = DataInputs(
            signal_image.data,
            options["background_image"].data,
            voxel_size_z,
            voxel_size_y,
            voxel_size_x,
        )

        detection_inputs = DetectionInputs(
            skip_detection,
            detected_cells,
            soma_diameter,
            ball_xy_size,
            ball_z_size,
            ball_overlap_fraction,
            log_sigma_size,
            n_sds_above_mean_thresh,
            soma_spread_factor,
            max_cluster_size,
        )

        if use_pre_trained_weights:
            trained_model = None
        classification_inputs = ClassificationInputs(
            skip_classification, use_pre_trained_weights, trained_model
        )

        if analyse_local:
            start_plane, end_plane = find_local_planes(
                options["viewer"], voxel_size_z, signal_image
            )
        elif not end_plane:
            end_plane = len(signal_image.data)

        misc_inputs = MiscInputs(
            start_plane, end_plane, n_free_cpus, analyse_local, debug
        )

        worker = Worker(
            data_inputs,
            detection_inputs,
            classification_inputs,
            misc_inputs,
        )

        worker.returned.connect(
            get_results_callback(skip_classification, options["viewer"])
        )
        # Make sure if the worker emits an error, it is propagated to this
        # thread
        worker.errored.connect(reraise)
        worker.connect_progress_bar_callback(progress_bar)

        worker.start()

    widget.native.layout().insertWidget(0, cellfinder_header())

    # reset restores defaults
    widget.reset_button.changed.connect(
        partial(restore_options_defaults, widget)
    )

    # Insert progress bar before the run and reset buttons
    widget.insert(widget.index("debug") + 1, progress_bar)

    # add the signal and background image etc.
    add_heavy_widgets(
        widget,
        (background_image_opt, signal_image_opt, cell_layer_opt),
        ("Background image", "Signal image", "Candidate cell layer"),
        ("voxel_size_z", "voxel_size_z", "soma_diameter"),
    )

    scroll = QScrollArea()
    scroll.setWidget(widget._widget._qwidget)
    widget._widget._qwidget = scroll

    return widget



import math
import multiprocessing.pool
import os
from functools import partial
from queue import Queue
from threading import Lock
from typing import Any, Callable, List, Optional, Tuple

import numpy as np
from brainglobe_utils.cells.cells import Cell
from tifffile import tifffile
from tqdm import tqdm

from cellfinder.core import logger
from cellfinder.core.detect.filters.setup_filters import (
    get_ball_filter,
    get_cell_detector,
)
from cellfinder.core.detect.filters.volume.structure_detection import (
    get_structure_centre,
)
from cellfinder.core.detect.filters.volume.structure_splitting import (
    StructureSplitException,
    split_cells,
)


class VolumeFilter(object):
    def __init__(
        self,
        *,
        soma_diameter: float,
        soma_size_spread_factor: float = 1.4,
        setup_params: Tuple[np.ndarray, Any, int, int, float, Any],
        n_planes: int,
        n_locks_release: int,
        save_planes: bool = False,
        plane_directory: Optional[str] = None,
        start_plane: int = 0,
        max_cluster_size: int = 5000,
        outlier_keep: bool = False,
        artifact_keep: bool = True,
    ):
        self.soma_diameter = soma_diameter
        self.soma_size_spread_factor = soma_size_spread_factor
        self.n_planes = n_planes
        self.z = start_plane
        self.save_planes = save_planes
        self.plane_directory = plane_directory
        self.max_cluster_size = max_cluster_size
        self.outlier_keep = outlier_keep
        self.n_locks_release = n_locks_release

        self.artifact_keep = artifact_keep

        self.clipping_val = None
        self.threshold_value = None
        self.setup_params = setup_params

        self.previous_plane: Optional[np.ndarray] = None

        self.ball_filter = get_ball_filter(
            plane=self.setup_params[0],
            soma_diameter=self.setup_params[1],
            ball_xy_size=self.setup_params[2],
            ball_z_size=self.setup_params[3],
            ball_overlap_fraction=self.setup_params[4],
        )

        self.cell_detector = get_cell_detector(
            plane_shape=self.setup_params[0].shape,  # type: ignore
            ball_z_size=self.setup_params[3],
            z_offset=self.setup_params[5],
        )

    def process(
        self,
        async_result_queue: Queue,
        locks: List[Lock],
        *,
        callback: Callable[[int], None],
    ) -> None:
        progress_bar = tqdm(total=self.n_planes, desc="Processing planes")
        for z in range(self.n_planes):
            # Get result from the queue.
            #
            # It is important to remove the result from the queue here
            # to free up memory once this plane has been processed by
            # the 3D filter here
            logger.debug(f"🏐 Waiting for plane {z}")
            result = async_result_queue.get()
            # .get() blocks until the result is available
            plane, mask = result.get()
            logger.debug(f"🏐 Got plane {z}")

            self.ball_filter.append(plane, mask)

            if self.ball_filter.ready:
                # Let the next 2D filter run
                z_release = z + self.n_locks_release + 1
                if z_release < len(locks):
                    logger.debug(f"🔓 Releasing lock for plane {z_release}")
                    locks[z_release].release()

                self._run_filter()

            callback(self.z)
            self.z += 1
            progress_bar.update()

        progress_bar.close()
        logger.debug("3D filter done")

    def _run_filter(self) -> None:
        logger.debug(f"🏐 Ball filtering plane {self.z}")
        # filtering original images, the images should be large enough in x/y
        # to benefit from parallelization. Note: don't pass arg as keyword arg
        # because numba gets stuck (probably b/c class jit is new)
        self.ball_filter.walk(True)

        middle_plane = self.ball_filter.get_middle_plane()
        if self.save_planes:
            self.save_plane(middle_plane)

        logger.debug(f"🏫 Detecting structures for plane {self.z}")
        self.previous_plane = self.cell_detector.process(
            middle_plane, self.previous_plane
        )

        logger.debug(f"🏫 Structures done for plane {self.z}")

    def save_plane(self, plane: np.ndarray) -> None:
        if self.plane_directory is None:
            raise ValueError(
                "plane_directory must be set to save planes to file"
            )
        plane_name = f"plane_{str(self.z).zfill(4)}.tif"
        f_path = os.path.join(self.plane_directory, plane_name)
        tifffile.imsave(f_path, plane.T)

    def get_results(self, worker_pool: multiprocessing.Pool) -> List[Cell]:
        logger.info("Splitting cell clusters and writing results")

        max_cell_volume = sphere_volume(
            self.soma_size_spread_factor * self.soma_diameter / 2
        )

        cells = []
        needs_split = []
        structures = self.cell_detector.get_structures().items()
        logger.debug(f"Processing {len(structures)} found cells")

        # first get all the cells that are not clusters
        for cell_id, cell_points in structures:
            cell_volume = len(cell_points)

            if cell_volume < max_cell_volume:
                cell_centre = get_structure_centre(cell_points)
                cells.append(Cell(cell_centre.tolist(), Cell.UNKNOWN))
            else:
                if cell_volume < self.max_cluster_size:
                    needs_split.append((cell_id, cell_points))
                else:
                    cell_centre = get_structure_centre(cell_points)
                    cells.append(Cell(cell_centre.tolist(), Cell.ARTIFACT))

        if not needs_split:
            logger.debug("Finished splitting cell clusters - none found")
            return cells

        # now split clusters into cells
        logger.debug(f"Splitting {len(needs_split)} clusters")
        progress_bar = tqdm(
            total=len(needs_split), desc="Splitting cell clusters"
        )

        # we are not returning Cell instances from func because it'd be pickled
        # by multiprocess which slows it down
        func = partial(_split_cells, outlier_keep=self.outlier_keep)
        for cell_centres in worker_pool.imap_unordered(func, needs_split):
            for cell_centre in cell_centres:
                cells.append(Cell(cell_centre.tolist(), Cell.UNKNOWN))
            progress_bar.update()

        progress_bar.close()
        logger.debug(
            f"Finished splitting cell clusters. Found {len(cells)} total cells"
        )

        return cells


def _split_cells(arg, outlier_keep):
    cell_id, cell_points = arg
    try:
        return split_cells(cell_points, outlier_keep=outlier_keep)
    except (ValueError, AssertionError) as err:
        raise StructureSplitException(f"Cell {cell_id}, error; {err}")


def sphere_volume(radius: float) -> float:
    return (4 / 3) * math.pi * radius**3



import os
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

# Check cellfinder is installed
try:
    __version__ = version("cellfinder")
except PackageNotFoundError as e:
    raise PackageNotFoundError("cellfinder package not installed") from e

# If Keras is not present, tools cannot be used.
# Throw an error in this case to prevent invocation of functions.
try:
    KERAS_VERSION = version("keras")
except PackageNotFoundError as e:
    raise PackageNotFoundError(
        f"cellfinder tools cannot be invoked without Keras. "
        f"Please install Keras with a backend into your environment "
        f"to use cellfinder tools. "
        f"For more information on Keras backends, please see "
        f"https://keras.io/getting_started/#installing-keras-3."
        f"For more information on brainglobe, please see "
        f"https://github.com/brainglobe/brainglobe-meta#readme."
    ) from e


# Set the Keras backend to torch
os.environ["KERAS_BACKEND"] = "torch"

__author__ = "Adam Tyson, Christian Niedworok, Charly Rousseau"
__license__ = "BSD-3-Clause"

DEFAULT_CELLFINDER_DIRECTORY = Path.home() / ".brainglobe" / "cellfinder"



from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from pathlib import Path

from cellfinder.core.download.download import (
    DEFAULT_DOWNLOAD_DIRECTORY,
    amend_user_configuration,
    download_models,
)


def download_parser(parser: ArgumentParser) -> ArgumentParser:
    """
    Configure the argument parser for downloading files.

    Parameters
    ----------
    parser : ArgumentParser
        The argument parser to configure.

    Returns
    -------
    ArgumentParser
        The configured argument parser.

    """

    parser.add_argument(
        "--install-path",
        dest="install_path",
        type=Path,
        default=DEFAULT_DOWNLOAD_DIRECTORY,
        help="The path to install files to.",
    )
    parser.add_argument(
        "--no-amend-config",
        dest="no_amend_config",
        action="store_true",
        help="Don't amend the config file",
    )
    parser.add_argument(
        "--model",
        dest="model",
        type=str,
        default="resnet50_tv",
        help="The model to use",
    )
    return parser


def get_parser() -> ArgumentParser:
    """
    Create an argument parser for downloading files.

    Returns
    -------
    ArgumentParser
        The configured argument parser.

    """
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = download_parser(parser)
    return parser


def main() -> None:
    """
    Run the main download function, and optionally amend the user
    configuration.

    """
    args = get_parser().parse_args()
    model_path = download_models(args.model, args.install_path)

    if not args.no_amend_config:
        amend_user_configuration(new_model_path=model_path)


if __name__ == "__main__":
    main()



import os
from typing import Callable, List, Optional, Tuple

import numpy as np
from brainglobe_utils.cells.cells import Cell

from cellfinder.core import logger
from cellfinder.core.download.download import model_type
from cellfinder.core.train.train_yml import depth_type


def main(
    signal_array: np.ndarray,
    background_array: np.ndarray,
    voxel_sizes: Tuple[int, int, int],
    start_plane: int = 0,
    end_plane: int = -1,
    trained_model: Optional[os.PathLike] = None,
    model_weights: Optional[os.PathLike] = None,
    model: model_type = "resnet50_tv",
    batch_size: int = 32,
    n_free_cpus: int = 2,
    network_voxel_sizes: Tuple[int, int, int] = (5, 1, 1),
    soma_diameter: int = 16,
    ball_xy_size: int = 6,
    ball_z_size: int = 15,
    ball_overlap_fraction: float = 0.6,
    log_sigma_size: float = 0.2,
    n_sds_above_mean_thresh: int = 10,
    soma_spread_factor: float = 1.4,
    max_cluster_size: int = 100000,
    cube_width: int = 50,
    cube_height: int = 50,
    cube_depth: int = 20,
    network_depth: depth_type = "50",
    skip_detection: bool = False,
    skip_classification: bool = False,
    detected_cells: List[Cell] = None,
    *,
    detect_callback: Optional[Callable[[int], None]] = None,
    classify_callback: Optional[Callable[[int], None]] = None,
    detect_finished_callback: Optional[Callable[[list], None]] = None,
) -> List:
    """
    Parameters
    ----------
    detect_callback : Callable[int], optional
        Called every time a plane has finished being processed during the
        detection stage. Called with the plane number that has finished.
    classify_callback : Callable[int], optional
        Called every time a point has finished being classified.
        Called with the batch number that has just finished.
    detect_finished_callback : Callable[list], optional
        Called after detection is finished with the list of detected points.
    """
    from cellfinder.core.classify import classify
    from cellfinder.core.detect import detect
    from cellfinder.core.tools import prep

    if not skip_detection:
        logger.info("Detecting cell candidates")

        points = detect.main(
            signal_array,
            start_plane,
            end_plane,
            voxel_sizes,
            soma_diameter,
            max_cluster_size,
            ball_xy_size,
            ball_z_size,
            ball_overlap_fraction,
            soma_spread_factor,
            n_free_cpus,
            log_sigma_size,
            n_sds_above_mean_thresh,
            callback=detect_callback,
        )

        if detect_finished_callback is not None:
            detect_finished_callback(points)
    else:
        points = detected_cells or []  # if None
        detect_finished_callback(points)

    if not skip_classification:
        install_path = None
        model_weights = prep.prep_model_weights(
            model_weights, install_path, model
        )
        if len(points) > 0:
            logger.info("Running classification")
            points = classify.main(
                points,
                signal_array,
                background_array,
                n_free_cpus,
                voxel_sizes,
                network_voxel_sizes,
                batch_size,
                cube_height,
                cube_width,
                cube_depth,
                trained_model,
                model_weights,
                network_depth,
                callback=classify_callback,
            )
        else:
            logger.info("No candidates, skipping classification")
    return points



from typing import List, Tuple

import napari
import napari.layers
import numpy as np
from brainglobe_utils.cells.cells import Cell
from brainglobe_utils.qtpy.logo import header_widget


def html_label_widget(label: str, *, tag: str = "b") -> dict:
    """
    Create a HMTL label for use with magicgui.
    """
    return dict(
        widget_type="Label",
        label=f"<{tag}>{label}</{tag}>",
    )


def cellfinder_header():
    """
    Create the header containing the brainglobe logo and documentation links
    for all cellfinder widgets.
    """
    return header_widget(
        "cellfinder",
        "Efficient cell detection in large images.",
        documentation_path="cellfinder/user-guide/napari-plugin/index.html",
        citation_doi="https://doi.org/10.1371/journal.pcbi.1009074",
        help_text="For help, hover the cursor over each parameter.",
    )


# the xyz axis order in napari relative to ours. I.e. our zeroth axis is the
# napari last axis. Ours is XYZ.
napari_points_axis_order = 2, 1, 0
# the xyz axis order in brainglobe relative to napari. I.e. napari's zeroth
# axis is our last axis - it's just flipped
brainglobe_points_axis_order = napari_points_axis_order


def add_classified_layers(
    points: List[Cell],
    viewer: napari.Viewer,
    unknown_name: str = "Rejected",
    cell_name: str = "Detected",
) -> None:
    """
    Adds cell candidates as two separate point layers - unknowns and cells, to
    the napari viewer. Does not add any other cell types, only Cell.UNKNOWN
    and Cell.CELL from the list of cells.
    """
    viewer.add_points(
        cells_to_array(points, Cell.UNKNOWN, napari_order=True),
        name=unknown_name,
        size=15,
        n_dimensional=True,
        opacity=0.6,
        symbol="ring",
        face_color="lightskyblue",
        visible=False,
        metadata=dict(point_type=Cell.UNKNOWN),
    )
    viewer.add_points(
        cells_to_array(points, Cell.CELL, napari_order=True),
        name=cell_name,
        size=15,
        n_dimensional=True,
        opacity=0.6,
        symbol="ring",
        face_color="lightgoldenrodyellow",
        metadata=dict(point_type=Cell.CELL),
    )


def add_single_layer(
    points: List[Cell],
    viewer: napari.Viewer,
    name: str,
    cell_type: int,
) -> None:
    """
    Adds all cells of cell_type Cell.TYPE to a new point layer in the napari
    viewer, with given name.
    """
    viewer.add_points(
        cells_to_array(points, cell_type, napari_order=True),
        name=name,
        size=15,
        n_dimensional=True,
        opacity=0.6,
        symbol="ring",
        face_color="lightskyblue",
        visible=True,
        metadata=dict(point_type=cell_type),
    )


def cells_to_array(
    cells: List[Cell], cell_type: int, napari_order: bool = True
) -> np.ndarray:
    """
    Converts all the cells of the given type as a 2D pos array.
    The column order is either XYZ, otherwise it's the napari ordering
    of the 3 axes (napari_points_axis_order).
    """
    cells = [c for c in cells if c.type == cell_type]
    if not cells:
        # make sure we return 2d array if cells is empty
        return np.zeros((0, 3), dtype=np.int_)
    points = np.array([(c.x, c.y, c.z) for c in cells])

    if napari_order:
        return points[:, napari_points_axis_order]
    return points


def napari_array_to_cells(
    points: napari.layers.Points,
    cell_type: int,
    brainglobe_order: Tuple[int, int, int] = brainglobe_points_axis_order,
) -> List[Cell]:
    """
    Takes a napari Points layer and returns a list of cell objects, one for
    each point in the layer.
    """
    data = np.asarray(points.data)[:, brainglobe_order].tolist()

    cells = []
    for row in data:
        cells.append(Cell(pos=row, cell_type=cell_type))

    return cells



import os
from collections.abc import Sequence
from typing import List, Optional, Tuple, Union

import keras
import numpy as np
from keras import Model

from cellfinder.core import logger
from cellfinder.core.classify.resnet import build_model, layer_type


def get_model(
    existing_model: Optional[os.PathLike] = None,
    model_weights: Optional[os.PathLike] = None,
    network_depth: Optional[layer_type] = None,
    learning_rate: float = 0.0001,
    inference: bool = False,
    continue_training: bool = False,
) -> Model:
    """Returns the correct model based on the arguments passed
    :param existing_model: An existing, trained model. This is returned if it
    exists
    :param model_weights: This file is used to set the model weights if it
    exists
    :param network_depth: This defines the type of model to be created if
    necessary
    :param learning_rate: For creating a new model
    :param inference: If True, will ensure that a trained model exists. E.g.
    by using the default one
    :param continue_training: If True, will ensure that a trained model
    exists. E.g. by using the default one
    :return: A keras model

    """
    if existing_model is not None or network_depth is None:
        logger.debug(f"Loading model: {existing_model}")
        return keras.models.load_model(existing_model)
    else:
        logger.debug(f"Creating a new instance of model: {network_depth}")
        model = build_model(
            network_depth=network_depth,
            learning_rate=learning_rate,
        )
        if inference or continue_training:
            logger.debug(
                f"Setting model weights according to: {model_weights}",
            )
            if model_weights is None:
                raise OSError("`model_weights` must be provided")
            model.load_weights(model_weights)
        return model


def make_lists(
    tiff_files: Sequence,
    train: bool = True,
) -> Union[Tuple[List, List], Tuple[List, List, np.ndarray]]:
    signal_list = []
    background_list = []
    if train:
        labels = []

    for group in tiff_files:
        for image in group:
            signal_list.append(image.img_files[0])
            background_list.append(image.img_files[1])
            if train:
                if image.label == "no_cell":
                    labels.append(0)
                elif image.label == "cell":
                    labels.append(1)

    if train:
        return signal_list, background_list, np.array(labels)
    else:
        return signal_list, background_list



from functools import lru_cache

import numpy as np
from numba import njit, objmode, prange
from numba.core import types
from numba.experimental import jitclass

from cellfinder.core.tools.array_operations import bin_mean_3d
from cellfinder.core.tools.geometry import make_sphere

DEBUG = False

uint32_3d_type = types.uint32[:, :, :]
bool_3d_type = types.bool_[:, :, :]
float_3d_type = types.float64[:, :, :]


@lru_cache(maxsize=50)
def get_kernel(ball_xy_size: int, ball_z_size: int) -> np.ndarray:
    # Create a spherical kernel.
    #
    # This is done by:
    # 1. Generating a binary sphere at a resolution *upscale_factor* larger
    #    than desired.
    # 2. Downscaling the binary sphere to get a 'fuzzy' sphere at the
    #    original intended scale
    upscale_factor: int = 7
    upscaled_kernel_shape = (
        upscale_factor * ball_xy_size,
        upscale_factor * ball_xy_size,
        upscale_factor * ball_z_size,
    )
    upscaled_ball_centre_position = (
        np.floor(upscaled_kernel_shape[0] / 2),
        np.floor(upscaled_kernel_shape[1] / 2),
        np.floor(upscaled_kernel_shape[2] / 2),
    )
    upscaled_ball_radius = upscaled_kernel_shape[0] / 2.0

    sphere_kernel = make_sphere(
        upscaled_kernel_shape,
        upscaled_ball_radius,
        upscaled_ball_centre_position,
    )
    sphere_kernel = sphere_kernel.astype(np.float64)
    kernel = bin_mean_3d(
        sphere_kernel,
        bin_height=upscale_factor,
        bin_width=upscale_factor,
        bin_depth=upscale_factor,
    )

    assert (
        kernel.shape[2] == ball_z_size
    ), "Kernel z dimension should be {}, got {}".format(
        ball_z_size, kernel.shape[2]
    )

    return kernel


# volume indices/size is 64 bit for very large brains(!)
spec = [
    ("ball_xy_size", types.uint32),
    ("ball_z_size", types.uint32),
    ("tile_step_width", types.uint64),
    ("tile_step_height", types.uint64),
    ("THRESHOLD_VALUE", types.uint32),
    ("SOMA_CENTRE_VALUE", types.uint32),
    ("overlap_fraction", types.float64),
    ("overlap_threshold", types.float64),
    ("middle_z_idx", types.uint32),
    ("_num_z_added", types.uint32),
    ("kernel", float_3d_type),
    ("volume", uint32_3d_type),
    ("inside_brain_tiles", bool_3d_type),
]


@jitclass(spec=spec)
class BallFilter:
    """
    A 3D ball filter.

    This runs a spherical kernel across the (x, y) dimensions
    of a *ball_z_size* stack of planes, and marks pixels in the middle
    plane of the stack that have a high enough intensity within the
    spherical kernel.
    """

    def __init__(
        self,
        plane_width: int,
        plane_height: int,
        ball_xy_size: int,
        ball_z_size: int,
        overlap_fraction: float,
        tile_step_width: int,
        tile_step_height: int,
        threshold_value: int,
        soma_centre_value: int,
    ):
        """
        Parameters
        ----------
        plane_width, plane_height :
            Width/height of the planes.
        ball_xy_size :
            Diameter of the spherical kernel in the x/y dimensions.
        ball_z_size :
            Diameter of the spherical kernel in the z dimension.
            Equal to the number of planes that stacked to filter
            the central plane of the stack.
        overlap_fraction :
            The fraction of pixels within the spherical kernel that
            have to be over *threshold_value* for a pixel to be marked
            as having a high intensity.
        tile_step_width, tile_step_height :
            Width/height of individual tiles in the mask generated by
            2D filtering.
        threshold_value :
            Value above which an individual pixel is considered to have
            a high intensity.
        soma_centre_value :
            Value used to mark pixels with a high enough intensity.
        """
        self.ball_xy_size = ball_xy_size
        self.ball_z_size = ball_z_size
        self.overlap_fraction = overlap_fraction
        self.tile_step_width = tile_step_width
        self.tile_step_height = tile_step_height

        self.THRESHOLD_VALUE = threshold_value
        self.SOMA_CENTRE_VALUE = soma_centre_value

        # getting kernel is not jitted
        with objmode(kernel=float_3d_type):
            kernel = get_kernel(ball_xy_size, ball_z_size)
        self.kernel = kernel

        self.overlap_threshold = np.sum(self.overlap_fraction * self.kernel)

        # Stores the current planes that are being filtered
        # first axis is z for faster rotating the z-axis
        self.volume = np.empty(
            (ball_z_size, plane_width, plane_height),
            dtype=np.uint32,
        )
        # Index of the middle plane in the volume
        self.middle_z_idx = int(np.floor(ball_z_size / 2))
        self._num_z_added = 0

        # first axis is z
        self.inside_brain_tiles = np.empty(
            (
                ball_z_size,
                int(np.ceil(plane_width / tile_step_width)),
                int(np.ceil(plane_height / tile_step_height)),
            ),
            dtype=np.bool_,
        )

    @property
    def ready(self) -> bool:
        """
        Return `True` if enough planes have been appended to run the filter.
        """
        return self._num_z_added >= self.ball_z_size

    def append(self, plane: np.ndarray, mask: np.ndarray) -> None:
        """
        Add a new 2D plane to the filter.
        """
        if DEBUG:
            assert [e for e in plane.shape[:2]] == [
                e for e in self.volume.shape[1:]
            ], 'plane shape mismatch, expected "{}", got "{}"'.format(
                [e for e in self.volume.shape[1:]],
                [e for e in plane.shape[:2]],
            )
            assert [e for e in mask.shape[:2]] == [
                e for e in self.inside_brain_tiles.shape[1:]
            ], 'mask shape mismatch, expected"{}", got {}"'.format(
                [e for e in self.inside_brain_tiles.shape[1:]],
                [e for e in mask.shape[:2]],
            )

        if self.ready:
            # Shift everything down by one to make way for the new plane
            # this is faster than np.roll, especially with z-axis first
            self.volume[:-1, :, :] = self.volume[1:, :, :]
            self.inside_brain_tiles[:-1, :, :] = self.inside_brain_tiles[
                1:, :, :
            ]

        # index for *next* slice is num we added *so far* until max
        idx = min(self._num_z_added, self.ball_z_size - 1)
        self._num_z_added += 1

        # Add the new plane to the top of volume and inside_brain_tiles
        self.volume[idx, :, :] = plane
        self.inside_brain_tiles[idx, :, :] = mask

    def get_middle_plane(self) -> np.ndarray:
        """
        Get the plane in the middle of self.volume.
        """
        return self.volume[self.middle_z_idx, :, :].copy()

    def walk(self, parallel: bool = False) -> None:
        # **don't** pass parallel as keyword arg - numba struggles with it
        # Highly optimised because most time critical
        ball_radius = self.ball_xy_size // 2
        # Get extents of image that are covered by tiles
        tile_mask_covered_img_width = (
            self.inside_brain_tiles.shape[1] * self.tile_step_width
        )
        tile_mask_covered_img_height = (
            self.inside_brain_tiles.shape[2] * self.tile_step_height
        )
        # Get maximum offsets for the ball
        max_width = tile_mask_covered_img_width - self.ball_xy_size
        max_height = tile_mask_covered_img_height - self.ball_xy_size

        # we have to pass the raw volume so walk doesn't use its edits as it
        # processes the volume. self.volume is the one edited in place
        input_volume = self.volume.copy()

        if parallel:
            _walk_parallel(
                max_height,
                max_width,
                self.tile_step_width,
                self.tile_step_height,
                self.inside_brain_tiles,
                input_volume,
                self.volume,
                self.kernel,
                ball_radius,
                self.middle_z_idx,
                self.overlap_threshold,
                self.THRESHOLD_VALUE,
                self.SOMA_CENTRE_VALUE,
            )
        else:
            _walk_single(
                max_height,
                max_width,
                self.tile_step_width,
                self.tile_step_height,
                self.inside_brain_tiles,
                input_volume,
                self.volume,
                self.kernel,
                ball_radius,
                self.middle_z_idx,
                self.overlap_threshold,
                self.THRESHOLD_VALUE,
                self.SOMA_CENTRE_VALUE,
            )


@njit(cache=True)
def _cube_overlaps(
    volume: np.ndarray,
    x_start: int,
    x_end: int,
    y_start: int,
    y_end: int,
    overlap_threshold: float,
    threshold_value: int,
    kernel: np.ndarray,
) -> bool:  # Highly optimised because most time critical
    """
    For each pixel in cube in volume that is greater than THRESHOLD_VALUE, sum
    up the corresponding pixels in *kernel*. If the total is less than
    overlap_threshold, return False, otherwise return True.

    Halfway through scanning the z-planes, if the total overlap is
    less than 0.4 * overlap_threshold, this will return False early
    without scanning the second half of the z-planes.

    Parameters
    ----------
    volume :
        3D array.
    x_start, x_end, y_start, y_end :
        The start and end indices in volume that form the cube. End is
        exclusive
    overlap_threshold :
        Threshold above which to return True.
    threshold_value :
        Value above which a pixel is marked as being part of a cell.
    kernel :
        3D array, with the same shape as *cube* in the volume.
    """
    current_overlap_value = 0.0

    middle = np.floor(volume.shape[0] / 2) + 1
    halfway_overlap_thresh = (
        overlap_threshold * 0.4
    )  # FIXME: do not hard code value

    for z in range(volume.shape[0]):
        # TODO: OPTIMISE: step from middle to outer boundaries to check
        # more data first
        #
        # If halfway through the array, and the overlap value isn't more than
        # 0.4 * the overlap threshold, return
        if z == middle and current_overlap_value < halfway_overlap_thresh:
            return False  # DEBUG: optimisation attempt

        for y in range(y_start, y_end):
            for x in range(x_start, x_end):
                # includes self.SOMA_CENTRE_VALUE
                if volume[z, x, y] >= threshold_value:
                    # x/y must be shifted in kernel because we x/y is relative
                    # to the full volume, so shift it to relative to the cube
                    current_overlap_value += kernel[
                        x - x_start, y - y_start, z
                    ]

    return current_overlap_value > overlap_threshold


@njit
def _is_tile_to_check(
    x: int,
    y: int,
    middle_z: int,
    tile_step_width: int,
    tile_step_height: int,
    inside_brain_tiles: np.ndarray,
) -> bool:  # Highly optimised because most time critical
    """
    Check if the tile containing pixel (x, y) is a tile that needs checking.
    """
    x_in_mask = x // tile_step_width  # TEST: test bounds (-1 range)
    y_in_mask = y // tile_step_height  # TEST: test bounds (-1 range)
    return inside_brain_tiles[middle_z, x_in_mask, y_in_mask]


def _walk_base(
    max_height: int,
    max_width: int,
    tile_step_width: int,
    tile_step_height: int,
    inside_brain_tiles: np.ndarray,
    input_volume: np.ndarray,
    volume: np.ndarray,
    kernel: np.ndarray,
    ball_radius: int,
    middle_z: int,
    overlap_threshold: float,
    threshold_value: int,
    soma_centre_value: int,
) -> None:
    """
    Scan through *volume*, and mark pixels where there are enough surrounding
    pixels with high enough intensity.

    The surrounding area is defined by the *kernel*.

    Parameters
    ----------
    max_height, max_width :
        Maximum offsets for the ball filter.
    inside_brain_tiles :
        3d array containing information on whether a tile is
        inside the brain or not. Tiles outside the brain are skipped.
    input_volume :
        3D array containing the plane-filtered data passed to the function
        before walking. volume is edited in place, so this is the original
        volume to prevent the changes for some cubes affective other cubes
        during a single walk call.
    volume :
        3D array containing the plane-filtered data - edited in place.
    kernel :
        3D array
    ball_radius :
        Radius of the ball in the xy plane.
    soma_centre_value :
        Value that is used to mark pixels in *volume*.

    Notes
    -----
    Warning: modifies volume in place!
    """
    for y in prange(max_height):
        for x in prange(max_width):
            ball_centre_x = x + ball_radius
            ball_centre_y = y + ball_radius
            if _is_tile_to_check(
                ball_centre_x,
                ball_centre_y,
                middle_z,
                tile_step_width,
                tile_step_height,
                inside_brain_tiles,
            ):
                if _cube_overlaps(
                    input_volume,
                    x,
                    x + kernel.shape[0],
                    y,
                    y + kernel.shape[1],
                    overlap_threshold,
                    threshold_value,
                    kernel,
                ):
                    volume[middle_z, ball_centre_x, ball_centre_y] = (
                        soma_centre_value
                    )


_walk_parallel = njit(parallel=True)(_walk_base)
_walk_single = njit(parallel=False)(_walk_base)
